/* **************************************************************************

   It can also be used for coldplug: Just start it with "hotplug coldusb".

   Copyright:  Freecom Technology GmbH, Berlin, 2004
               www.freecom.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

 ************************************************************************* */

#include "freecom.h"

#include <dirent.h>
#include <syslog.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/mount.h>


// write errors in syslog file ?
#define ERROR_LOG
// write infos in syslog file ?
//#define INFO_LOG

// react on key-pressed events
#define KEY_ACTIONS


// where and how to mount mass storage devices
#define  MOUNT_DIR      "/home/"
#define  PRINT_COMMAND  "/sbin/p910nd -f /dev/usblp%s %s"

// possible actions
#define  REMOVE_DEVICE   0
#define  ADD_DEVICE      1
#define  COLDPLUG_DEVICE 2

// ----------------------------------------------------------------------
// tell Samba to re-read configuration
int reconfigSamba()
{
  int Result = -1;
  fflush(stdout);

  switch(fork()) {
    case -1:  break;
    case  0:  fclose(stdout);
              fclose(stderr);
              execl("/usr/bin/killall", "killall", "-SIGHUP", "smbd", 0);
              exit(1);
    default:  wait(&Result);
  }

  return Result;
}

// ----------------------------------------------------------------------
void switchOnUsbLed(char PortNo)
{
  #ifdef DEVICE_IS_DT2
  char *p;
  FILE *File;
  if(PortNo != '2')
    p = "/sys/class/leds/USB1/brightness";
  else
    p = "/sys/class/leds/USB2/brightness";
  File = fopen(p, "w");
  if(File) {
    fwrite("1", 1, sizeof(char), File);
    fclose(File);
  }
  #endif

  #ifdef DEVICE_IS_VT4
  if(PortNo != '2')
    system("/sbin/led USB1 1");
  else
    system("/sbin/led USB2 1");
  #endif

  #ifdef DEVICE_IS_FSG
  FILE *File;
  File = fopen("/sys/class/leds/USB/brightness", "w");
  if(File) {
    fwrite("1", 1, sizeof(char), File);   // LED on
    fclose(File);
  }
  #endif
}

// ----------------------------------------------------------------------
void switchOffUsbLed(char PortNo)
{
  #ifdef DEVICE_IS_DT2
  char *p;
  FILE *File;
  if(PortNo != '2')
    p = "/sys/class/leds/USB1/brightness";
  else
    p = "/sys/class/leds/USB2/brightness";
  File = fopen(p, "w");
  if(File) {
    fwrite("0", 1, sizeof(char), File);
    fclose(File);
  }
  #endif

  #ifdef DEVICE_IS_VT4
  if(PortNo != '2')
    system("/sbin/led USB1 0");
  else
    system("/sbin/led USB2 0");
  #endif

  #ifdef DEVICE_IS_FSG
  FILE *File;
  File = fopen("/sys/class/leds/USB/brightness", "w");
  if(File) {
    fwrite("0", 1, sizeof(char), File);   // LED on
    fclose(File);
  }
  #endif
}

// ----------------------------------------------------------------------
// Wait until all hotplug processes are terminated, that were started
// before the own process.
void waitForMyRound()
{
  int z, Count, Laps = 0;
  char Buffer[maxLen];
  DIR *pDir;
  struct dirent *procEntry;
  FILE *File;

  pid_t PID;
  pid_t myPID = getpid();

  do {
    pDir = opendir("/proc");
    if(pDir == 0)  return;

    // search for process of hotplugs
    Count = 0;
    while((procEntry=readdir(pDir)) != 0) {
      if(procEntry->d_name[0] > '9')  continue; // is process entry ?
      if(procEntry->d_name[0] < '0')  continue; // is dummy ?

      PID = atoi(procEntry->d_name);
      if(PID >= myPID)
        continue;

      sprintf(Buffer, "/proc/%s/cmdline", procEntry->d_name);
      File = fopen(Buffer,"r");
      if(File == 0)  continue;

      z = fread(Buffer, 1, maxLen-1, File);
      if(z < 2)  continue;
      Buffer[z] = 0;
      fclose(File);

      if(strcmp(Buffer, "/sbin/hotplug") != 0)
        continue;   // didn't found process searched for

      Count++;
      break;
    }
    closedir(pDir);

    if(Count > 0)
      sleep(1);

    Laps++;
  } while((Count > 0) && (Laps < 15));
}

// ----------------------------------------------------------------------
// Creates the name of the directory to which a SCSI device is
// mounted. It reads the product name and the serial number from
// the "SysPath" directory.
int createMountDirectoryName(char *SysPath, char *MountPath)
{
  #ifdef ERROR_LOG
    syslog(LOG_ERR, "Current mountpath %s\n", MountPath);
  #endif

  strcat(SysPath, "product");

  FILE *File = fopen(SysPath, "r");
  if(!File) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot get model from \"%s\".", SysPath);
    #endif
    return -1;
  }

  // first part of mount folder is the model name
  strcpy(MountPath, MOUNT_DIR);
  char *p = MountPath + strlen(MountPath);
  *(p + fread(p, 1, 200, File)) = 0;
  fclose(File);


  p = strrchr(SysPath, '/');
  if(p == 0)  return -1;
  strcpy(p+1, "serial");

	File = fopen(SysPath, "r");
	if(!File)
	{
		#ifdef ERROR_LOG
		syslog(LOG_ERR, "Cannot get serial number from \"%s\".", SysPath);
		#endif
		//return -1;

		// create fake serial number
		// just use the name, don't add anything
		//strcat(MountPath,"-1111");
	}
	else
	{
		p = MountPath + strlen(MountPath);
		*(p++) = '-';
		*(p + fread(p, 1, 4, File)) = 0; // second part of mount folder are ...
		fclose(File);                    // ... 4 characters of the S/N
	}

  // Replace forbidden characters in mount directory name with "_".
  p = MountPath + strlen(MOUNT_DIR) - 1;
  while(*(++p)) {
    if(*p == '-') continue;
    if((*p >=  '0') && (*p <= '9')) continue;
    if((*p >=  'A') && (*p <= 'Z')) continue;
    if((*p >=  'a') && (*p <= 'z')) continue;
    if(*p == '&') continue;
    *p = '_';
  }

  // MvdB
  // postprocess the name to see if there is a config for the mount name, 
  // so we can use a more sane name for that.
  // Since I am a Java weenie this code probably sucks, but it works ;)

  char *AliasConfFile = "/etc/drivealias.conf";
  char *AliasConfig = readfile(AliasConfFile,0);
  if (AliasConfig != NULL) {
    // try to get the correct value
    char *aliasValue = (char*)malloc(maxLen);
    char *tmp = (char*)malloc(maxLen);
    char *id = (char*)malloc(maxLen);
    strcpy(tmp, MountPath);
    strncpy(id, tmp+(strlen(MOUNT_DIR)), strlen(tmp)-strlen(MOUNT_DIR));
    if (get_value(AliasConfig, id, aliasValue) == 1) {
      strcpy(MountPath, MOUNT_DIR);
      strcat(MountPath, aliasValue);
      #ifdef ERROR_LOG
        syslog(LOG_ERR, "Changing mount directoryname to %s\n", aliasValue);
      #endif
    }
  }

  return 0;
}


// ***********************************************************************
// Handle USB mass-storage events.
// ***********************************************************************
int addMassStorageDevice(char *MountDir, char *Dev, char PortNo)
{
  char Buffer[maxLen], realDir[maxLen];
  FILE *File = fopen("/proc/mounts", "r");   // get mounted devices
  if(File == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot read \"/proc/mounts\".");
    #endif
    return -1;
  }
  Buffer[fread(Buffer, 1, maxLen-1, File)] = 0;
  fclose(File);

  if(!strstr(Buffer, " /home "))
    return -1;   // mount partition not yet ready (startup)

  sprintf(realDir, "\n%s ", Dev);
  if(strstr(Buffer, realDir))
    return -1;   // partition already mounted


  int Result;
  umask(0);
  // Create directory to mount to. Search new name, if it already exists.
  strcpy(realDir, MountDir);
  Result = mkdir(realDir, S_IRWXO | S_IRWXG | S_IRWXU);
  if(Result < 0) {
    if(errno != EEXIST)  return -1;
    Result = 0;
    do {
      if((Result++) > 30) return -1;
      sprintf(realDir, "%s-%d", MountDir, Result);
    } while(mkdir(realDir, S_IRWXO | S_IRWXG | S_IRWXU) < 0);
  }

  if(chown(realDir, 0, 105) != 0) {
    #ifdef ERROR_LOG
      syslog(LOG_WARNING, "Command \"%s\" not successful.", Buffer);
    #endif
  }


  // Finally, try to mount it.
  Result = -1;
  int Data = 0;
  if(Dev[6] == 'r')  // Is it a CDROM ?
    Result = mount(Dev, realDir, "iso9660",
        MS_MGC_VAL | MS_SYNCHRONOUS | MS_NOATIME | MS_NODIRATIME, &Data);

  else {
    // if it is not a CDROM -> test all supported filesystems

    // first try NTFS
    fflush(stdout);
    switch(fork()) {
      case -1:  break;
      case  0:  fclose(stdout);
                fclose(stderr);
                execl("/bin/ntfs-3g", "ntfs-3g", Dev, realDir, "-o", "force,locale=en_GB.UTF-8", 0);
                exit(1);
      default:  wait(&Result);
    }


    if(Result != 0) {
      // try all non-NTFS filesystems
      File = fopen("/proc/filesystems", "r");
      if(File == 0) {
        #ifdef ERROR_LOG
        syslog(LOG_ERR, "Cannot read \"/proc/filesystems\".");
        #endif
        return -1;
      }

      while(!feof(File)) {
        if(!fgets(Buffer, maxLen, File))   // read row by row
          break;
        if(Buffer[0] != '\t')
          continue;
        Buffer[strlen(Buffer)-1] = 0;  // cut off linefeed

        Result = mount(Dev, realDir, Buffer+1,
            MS_MGC_VAL | MS_SYNCHRONOUS | MS_NOATIME | MS_NODIRATIME, &Data);
        if(Result != 0)  continue;
        break;
      }
      fclose(File);
    }
  }


  if(Result != 0) {
    #ifdef ERROR_LOG
      syslog(LOG_WARNING, "Mounting \"%s\" to \"%s\" not successful: %d.", Dev, realDir, errno);
    #endif

    rmdir(realDir);    // remove directory if not mountable
    return Result;
  }



  // set USB LED on
  switchOnUsbLed(PortNo);


  // Create entry in configuration file.
  char *Users, *Config;
  Config = readfile(MasterConfName, 0);
  if(Config) {
    sprintf(Buffer, "\nusb=%s#", realDir+6);
    Users = strstr(Config, Buffer);
    if(Users == 0) {   // does already exist ?
      sprintf(Buffer, "\n[special]\nusb=%s#\n", realDir+6);
      Config = replace_string_once(Config, Config, "\n[special]\n", Buffer);
      Result = writefile(MasterConfName, Config);
      #ifdef ERROR_LOG
        if(Result < 0)
          syslog(LOG_ERR, "Cannot write \"%s\".", MasterConfName);
      #endif
    }
    else free(Config);
  }
  else {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot read \"%s\".", MasterConfName);
    #endif
    return -1;
  }

  return 0;
}

// ----------------------------------------------------------------------
// Mount all partitions of the device.
int addMassStorage(char *MountPath, char *DevPath, char PortNo)
{
  FILE *File = fopen("/proc/partitions", "r");
  if(!File) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot open \"/proc/partitions\".");
    #endif
    return 0;
  }

  char Buffer[maxLen], *p;
  fgets(Buffer, maxLen, File);  //  first row not needed
  fgets(Buffer, maxLen, File);  // second row not needed
  while(!feof(File)) {
    if(!fgets(Buffer, maxLen, File))  // read row by row
      break;
    Buffer[strlen(Buffer)-1] = 0;    // cut off linefeed
    p = strrchr(Buffer, ' ');
    if(!p++)  continue;

    if(strncmp(p, DevPath+5, 3) == 0) { // belongs partition to device ?
      strcpy(DevPath+5, p); // try also the device itself (partitionless)

      if(addMassStorageDevice(MountPath, DevPath, PortNo) == 0) // success ?
        if(DevPath[strlen(DevPath)-1] > '9') // partitionless device ?
          break;   // do not try the partitions, they are false
    }
  }
  fclose(File);

  return 1;
}

// ----------------------------------------------------------------------
int removeMassStorageDevice(char *Dev, char PortNo)
{
  int len, Result2=0, Result3=0;
  char Line[maxLen], *realDir=0, *p;

  // First, find mount directory.
  FILE *File = fopen("/proc/mounts", "r");   // get mounted devices
  if(!File) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot read \"/proc/mounts\".");
    #endif
    return -1;
  }

  len = strlen(Dev);
  while(!feof(File)) {
    if(!fgets(Line, maxLen, File))   // read row by row
      break;
    if(strncmp(Line, Dev, len) != 0)
      continue;
    if(Line[len] != ' ')
      continue;
    realDir = Line + len + 1;
    p = strchr(realDir, ' ');  // The folder name never contains spaces!
    if(!p) return -1;
    *p = 0;
    break;
  }
  fclose(File);


  if(realDir) {
    // umount directory and remove it.  ....................................
    Result2 = umount(realDir);
    #ifdef ERROR_LOG
    if(Result2 != 0)
      syslog(LOG_ERR, "Cannot unmount \"%s\": %d.", realDir, errno);
    #endif


    Result3 = rmdir(realDir);
    #ifdef ERROR_LOG
    if(Result3)
      syslog(LOG_ERR, "Command \"rmdir %s\" not successful: %d.", realDir, Result3);
    #endif
  }


  // set USB LED off
  switchOffUsbLed(PortNo);

  return Result2 | Result3;
}

// ----------------------------------------------------------------------
// Removes all SCSI devices except the root device.
int removeMassStorage(char *DevPath, char PortNo)
{
  int Result, len=0, Count=0;
  char Line[maxLen], *realDir=0, *p;

  // First, find mount directory.
  FILE *File = fopen("/proc/mounts", "r");   // get mounted devices
  if(!File) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot read \"/proc/mounts\".");
    #endif
    return 0;
  }

  if(DevPath)
    len = strlen(DevPath);

  while(!feof(File)) {
    if(!fgets(Line, maxLen, File))   // read row by row
      break;
    if(strncmp(Line, "/dev/sd", 7) != 0)
      if(strncmp(Line, "/dev/sr", 7) != 0)
        continue;
    #ifndef DEVICE_IS_VT4
    if(strncmp(Line, "/dev/sda", 8) == 0)
      continue;
    #endif

    if(DevPath)
      if(strncmp(Line, DevPath, len) != 0)
        continue;

    realDir = strchr(Line, ' ');
    if(realDir == 0)
      continue;
    realDir++;

    p = strchr(realDir, ' ');  // folder name never contains spaces
    if(!p) continue;
    *p = 0;

    Result = umount(realDir);
    if(Result == 0)
      Count++;
    #ifdef ERROR_LOG
    else
      syslog(LOG_ERR, "Cannot unmount \"%s\": %d.", realDir, errno);
    #endif

    Result = rmdir(realDir);
    #ifdef ERROR_LOG
    if(Result)
      syslog(LOG_ERR, "Command \"rmdir %s\" not successful: %d.", realDir, Result);
    #endif

  }

  fclose(File);


  if(Count > 0)
    switchOffUsbLed(PortNo);
  return Count;
}


// ***********************************************************************
// Handle USB printer events. Returns 1 if successful.
// ***********************************************************************
int addPrinter(char *No, char PortNo)
{
  char Command[maxLen];

  sprintf(Command, PRINT_COMMAND, No, No);
  int Result = system(Command);
  if(Result) {
    #ifdef ERROR_LOG
    syslog(LOG_ERR, "Command \"%s\" not successful.", Command);
    #endif
    return 0;
  }

  switchOnUsbLed(PortNo);
  return 1;
}

// -----------------------------------------------------------------------
// Terminate printer driver for printer number "No".
// If "No" == 0, then terminate all printer drivers.
int removePrinter(char *No, char PortNo)
{
  FILE *fp;
  DIR *pDir;
  struct dirent *procEntry;

  if((pDir=opendir("/proc")) == NULL) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot enter \"%s\".", "/proc");
    #endif
    return -1;
  }

  char Buffer[maxLen], Command[maxLen];
  if(No)
    sprintf(Command, "/sbin/p910%sd", No);
  else
    strcpy(Command, "/sbin/p910");

  // search for process of printer driver
  int Result;
  while((procEntry=readdir(pDir)) != 0) {
    if(procEntry->d_name[0] > '9')  continue;  // is process entry ?
    if(procEntry->d_name[0] < '0')  continue;  // is dummy ?

    sprintf(Buffer, "/proc/%s/cmdline", procEntry->d_name);
    fp = fopen(Buffer,"r");  // proc files have length 0 !!!
    if(!fp) {
      #ifdef ERROR_LOG
        syslog(LOG_ERR, "Cannot open \"%s\".", Buffer);
      #endif
      continue;
    }
    Result = fread(Buffer, 1, maxLen-1, fp);
    if(Result < 2)  continue;
    Buffer[Result] = 0;
    fclose(fp);

    if(No) {
      if(strcmp(Buffer, Command) != 0)
        continue;   // didn't found process searched for
    }
    else {
      if(strncmp(Buffer, Command, strlen(Command)) != 0)
        continue;   // didn't found process searched for
    }

    Result = atoi(procEntry->d_name);    // read PID
    if(Result < 1) {
      #ifdef ERROR_LOG
        syslog(LOG_ERR, "Wrong process PID \"%s\".", procEntry->d_name);
      #endif
      return -1;
    }

    Result = kill(Result, SIGTERM);
    #ifdef ERROR_LOG
      if(Result)
        syslog(LOG_ERR, "Command \"kill %s\" not successful.", procEntry->d_name);
    #endif

    if(No)
      switchOffUsbLed(PortNo);

    if(No) {
      closedir(pDir);
      return 0;
    }
  }


  if(No) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot find \"%s\" to remove.", Command);
    #endif
    closedir(pDir);
    return 1;
  }

  closedir(pDir);
  return 0;
}

// --------------------------------------------------------------------------
void hotpluggedPrinter()
{
  char *EnvSysPath = getenv("DEVPATH");
  if(EnvSysPath == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Missing device path.");
    #endif
    return;
  }

  char *p = strrchr(EnvSysPath, '/');
  if(p == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Wrong device path.");
    #endif
    return;
  }

  #ifdef INFO_LOG
    syslog(LOG_NOTICE, "Device \"%s\".", p);
  #endif

  if(strncmp(p, "/lp", 3) != 0)    // is it a printer ?
    return;

  char *No = p + 3;

  // .............................................................
  p = getenv("ACTION");
  if(p == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "%s.", "Missing action statement");
    #endif
    return;
  }

  EnvSysPath = getenv("PHYSDEVPATH");
  if(EnvSysPath == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Missing physical device path.");
    #endif
    return;
  }

  #ifdef DEVICE_IS_DT2
  EnvSysPath = strstr(EnvSysPath, "/usb");
  if(EnvSysPath == 0)  return;   // it is not a USB device !

  FILE *File;
  if(*(EnvSysPath+4) != '2')
    File = fopen("/sys/class/leds/USB1/brightness", "w");
  else
    File = fopen("/sys/class/leds/USB2/brightness", "w");

  if (File) {
    fwrite ("2", 1, sizeof(char), File);
    fclose (File);
  }
  #endif
  #ifdef DEVICE_IS_VT4
  EnvSysPath = strstr(EnvSysPath, "/usb");
  if(EnvSysPath == 0)  return;   // it is not a USB device !
  #endif

  #ifdef INFO_LOG
    syslog(LOG_NOTICE, "EnvSysPath \"%s\".", EnvSysPath);
  #endif

  if(strcmp(p, "add") == 0)
    addPrinter(No, *(EnvSysPath+4));
  else if(strcmp(p, "remove") == 0)
    removePrinter(No, *(EnvSysPath+4));
  else {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Unknown action \"%s\".", p);
    #endif
    switchOffUsbLed(*(EnvSysPath+4));
    return;
  }
}


// ***********************************************************************
// Remove all directories in /home that are not used but empty.
// ***********************************************************************
void cleanupFolders()
{
  char Buffer[maxLen];

  char *content = readfile(MasterConfName, 0);
  if(!content) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot read \"%s\".", MasterConfName);
    #endif
    return;
  }

  struct dirent *homeEntry;
  DIR *pDir = opendir(MOUNT_DIR);
  if(!pDir) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot open \"%s\".", MOUNT_DIR);
    #endif
    return;
  }

  int Result;
  readdir(pDir);   // ignore "."
  readdir(pDir);   // ignore ".."

  // Delete all directories that are not shared folfers.
  while((homeEntry=readdir(pDir)) != 0) {

    sprintf(Buffer, "\ndir=%s#", homeEntry->d_name);
    if(strstr(content, Buffer))
      continue;

    sprintf(Buffer, "/home/%s", homeEntry->d_name);
    Result = rmdir(Buffer);    // remove directory if empty
    #ifdef ERROR_LOG
      if(Result)  syslog(LOG_ERR, "Command \"rmdir %s\" not successful.",
                                  Buffer);
    #endif
  }
  closedir(pDir); // not needed anymore ("procEntry->d_name" now invalid!)
  free(content);

  return;
}


#ifdef DEVICE_IS_FSG
// ***********************************************************************
// Handle SATA harddrives. Returns 1 if device present.
// ***********************************************************************
int mountSATA()
{
  char MountDir[128], Command[maxLen], Buffer[maxLen], *p;

  FILE *File = fopen("/sys/block/sdb/device/scsi_device:2:0:0:0", "r");
  if(!File)
    return 0;   // return, if no external SATA drive connected
  fclose(File);


  File = fopen("/proc/partitions", "r");
  if(!File) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot open \"/proc/partitions\".");
    #endif
    return 0;
  }

  char *Config, *DirName;
  Config = readfile(MasterConfName, 0);
  if(!Config) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot read \"%s\".", MasterConfName);
    #endif
  }


  int modifiedConf = 0;
  while(!feof(File)) {
    if(fgets(Buffer, maxLen, File) == 0) break;   // read row by row
    Buffer[strlen(Buffer)-1] = 0;  // remove linefeed
    p = strstr(Buffer, " sdb");
    if(!p)  continue;

    sprintf(MountDir, "/home/external_SATA_drive%s", p+4);
    umask(0);
    if(mkdir(MountDir, S_IRWXO | S_IRWXG | S_IRWXU) != 0) {
      #ifdef ERROR_LOG
        syslog(LOG_WARNING, "Command \"mkdir %s\" not successful.", MountDir);
      #endif
      continue;
    }

#define  MOUNT_COMMAND  "/bin/mount -o noatime,nodiratime "
    strcpy(Command, MOUNT_COMMAND);
    strcat(Command, "/dev/");
    strcat(Command, p+1);
    strcat(Command, " ");
    strcat(Command, MountDir);
    if(system(Command) != 0) {
      #ifdef ERROR_LOG
        syslog(LOG_WARNING, "Command \"%s\" not successful.", Command);
      #endif
      rmdir(MountDir);    // remove directory if not mountable
      continue;
    }


    p = strchr(MountDir+1, '/');  // cut off "/home/"
    if(p)
      DirName = p + 1;
    else
      DirName = MountDir;

    // Create entry in master.conf if not already there.
    if(Config) {
      sprintf(Command, "\nusb=%s#", DirName);
      p = strstr(Config, Command);
      if(p == 0) {
        sprintf(Command, "\n[special]\nusb=%s#\n", DirName);
        Config = replace_string_once(Config, Config, "\n[special]\n", Command);
        modifiedConf = 1;
      }
    }

    FILE *led = fopen("/sys/class/leds/SATA/brightness", "w");
    if (led) {
      fwrite ("1", 1, sizeof(char), led);
      fclose (led);
    }
  }

  fclose(File);


  int Result;
  if(modifiedConf) {
    Result = writefile(MasterConfName, Config);
    #ifdef ERROR_LOG
      if(Result < 0)
        syslog(LOG_ERR, "Cannot write \"%s\".", MasterConfName);
    #endif
  }
  else
    free(Config);

  return 1;
}
#endif    // of "DEVICE_IS_FSG"


// ***********************************************************************
// Handle cold-plug event.
// ***********************************************************************
int coldpluggedDevice(char PortNo, int Action)
{
  int No, Count = 0;
  DIR *sysDir, *scsiDir;
  struct dirent *sysEntry;
  char *p, currPort;
  char DevPath[16];
  char MountPath[256];
  char SysPath[maxLen] = "/sys/bus/usb/drivers/usb-storage";
  char realPortNo;

  // ......................................................
  // coldplug events need to work on every mass storage device!
  sysDir = opendir(SysPath);
  if(!sysDir) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot enter \"%s\".", SysPath);
    #endif
    return -1;
  }

  strcat(SysPath, "/bla/host0/");

  // Search for USB mass storage devices.
  while((sysEntry=readdir(sysDir)) != 0) {
    if(*(sysEntry->d_name) < '0')
      continue;
    if(*(sysEntry->d_name) > '9')
      continue;
    currPort = *(sysEntry->d_name);
    if(PortNo > '0')
      if(currPort != PortNo)
        continue;

    p = strstr(SysPath, "/host");
    if(p == 0)
      continue;
    *p = 0;

    p = strrchr(SysPath, '/');
    if(p == 0)
      continue;

    strcpy(p+1, sysEntry->d_name);
    strcat(SysPath, "/../");
    if(createMountDirectoryName(SysPath, MountPath) < 0)
      continue;


    p = strstr(SysPath, "/../");
    if(p == 0)
      continue;
    *p = 0;


    // Get "host" directory.
    scsiDir = opendir(SysPath);
    if(!scsiDir) {
      #ifdef ERROR_LOG
        syslog(LOG_ERR, "Cannot enter \"%s\".", SysPath);
      #endif
      continue;
    }

    while((sysEntry=readdir(scsiDir)) != 0) {
      if(strncmp(sysEntry->d_name, "host", 4) == 0)
        break;
    }
    if(sysEntry == 0) {
      closedir(scsiDir);
      continue;
    }

    No = atoi(sysEntry->d_name + 4);
    closedir(scsiDir);
    sprintf(SysPath+strlen(SysPath),
        "/host%d/target%d:0:0/%d:0:0:0", No, No, No);


    // Get "block" directory to obtain device name.
    #ifdef DEVICE_IS_FSG
    scsiDir = opendir(SysPath);
    if(!scsiDir) {
      #ifdef ERROR_LOG
        syslog(LOG_ERR, "Cannot enter \"%s\".", SysPath);
      #endif
      continue;
    }

    while((sysEntry=readdir(scsiDir)) != 0) {
      if(strncmp(sysEntry->d_name, "block:", 6) == 0)
        break;
    }

    if(sysEntry != 0)
      sprintf(DevPath, "/dev/%s", sysEntry->d_name + 6);
    closedir(scsiDir);

    if(sysEntry == 0)
      continue;
    #endif

    #ifndef DEVICE_IS_FSG
    char Buffer[256];
    strcat(SysPath, "/block");
    No = readlink(SysPath, Buffer, 256);
    if(No < 0)
      continue;
    Buffer[No] = 0;
    sprintf(DevPath, "/dev/%s", Buffer + No - 3);
    #endif

    // Mount all partitions of the device.
    if(Action == REMOVE_DEVICE)
      Count += removeMassStorage(DevPath, currPort);
    else
      Count += addMassStorage(MountPath, DevPath, currPort);
  }

  closedir(sysDir);

  if(PortNo > '0')
    if(Count > 0) {
      reconfigSamba();
      return Count;  // work is done!
    }

  #ifdef INFO_LOG
    syslog(LOG_NOTICE, "PortNo \"%c\".", PortNo);
  #endif

  // ......................................................
  // coldplug events for every printer device
  if((sysDir=opendir("/sys/class/usb")) == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Cannot enter \"/sys/class/usb\".");
    #endif
  }
  else {

    while((sysEntry=readdir(sysDir)) != 0)
      if(strncmp(sysEntry->d_name, "lp", 2) == 0) {
        if(PortNo > '0')
          if(*(sysEntry->d_name+2) != PortNo)
            continue;

			realPortNo=*(sysEntry->d_name+2);

			// get real port number
			#ifndef DEVICE_IS_FSG
				if(PortNo < '0')
				{
					char Buffer[256];
					sprintf(MountPath, "/sys/class/usb/%s/device",sysEntry->d_name);
					No = readlink(MountPath, Buffer, 256);
					if(No > 0)
					{
						Buffer[No] = 0;
						if ((p=strstr(Buffer,"/usb"))!=NULL)
							realPortNo=p[4];
					}
				}
			#endif

        if(Action == REMOVE_DEVICE)
          removePrinter(sysEntry->d_name + 2,realPortNo);
        else
          addPrinter(sysEntry->d_name + 2,realPortNo);
      }

    closedir(sysDir);
  }


  // ......................................................
  // coldplug events for external SATA device
  #ifdef DEVICE_IS_FSG
  if(Action == ADD_DEVICE)
    Count += mountSATA();
  #endif


  if(Count > 0)
    reconfigSamba();
  return Count;
}


// ***********************************************************************
// Handle hotplugged SCSI device.
// ***********************************************************************
void hotpluggedBlockDevice()
{
  char *EnvSysPath = getenv("DEVPATH");
  if(EnvSysPath == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Missing device path.");
    #endif
    return;
  }

  char *p = strrchr(EnvSysPath, '/');
  if(p == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Wrong device path.");
    #endif
    return;
  }

  char DevPath[16] = "/dev/";   // <<<<<<<<<<<< device to be mounted
  strcat(DevPath, p+1);


  // .............................................................
  p = getenv("ACTION");
  if(p == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "%s.", "Missing action statement");
    #endif
    return;
  }

  int Action;
  if(strcmp(p, "add") == 0)
    Action = ADD_DEVICE;
  else if(strcmp(p, "remove") == 0)
    Action = REMOVE_DEVICE;
  else {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Unknown action \"%s\".", p);
    #endif
    return;
  }


  // .............................................................
  if(Action == ADD_DEVICE) {
    // In order to preserv the partition order, we wait until the
    // former called hotplug processes are terminated.
    waitForMyRound();
  }

  // .............................................................
  EnvSysPath = getenv("PHYSDEVPATH");
  if(EnvSysPath == 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Missing physical device path.");
    #endif
    return;
  }

  char SysPath[1024] = "/sys";
  strcat(SysPath, EnvSysPath);
  p = strstr(SysPath, "/usb");
  if(p == 0)  return;   // it is not a USB device !
  p += 4;
  char PortNo = *p;
  p = strchr(p, '/');
  if(p == 0)  return;   // it is not a USB device !
  p = strchr(p+4, '/');
  if(p == 0)  return;   // it is not a USB device !
  *(p+1) = 0;


  char MountPath[256];     // <<<<<<<<<<<< where to mount it
  if(createMountDirectoryName(SysPath, MountPath) < 0)
    return;


  #ifdef DEVICE_IS_DT2
  FILE *File;
  if(PortNo != '2')
    p = "/sys/class/leds/USB1/brightness";
  else
    p = "/sys/class/leds/USB2/brightness";
  File = fopen(p, "w");
  if(File) {
    fwrite("2", 1, sizeof(char), File);   // set LED to blinking
    fclose(File);
  }
  #endif

  #ifdef DEVICE_IS_FSG
  // ..................................................................
  // After a "save unplug" the USB LED shines. If a user plugs in a
  // new device all devices should be re-pluged.
  char Brightness[4];
  FILE *File = fopen("/sys/class/leds/USB/brightness", "r");
  if(File == 0)  return;
  Brightness[0] = 0;
  Brightness[fread(Brightness, 3, sizeof(char), File)] = 0;
  fclose(File);

  if(Brightness[0] > '0') {
    if(Action == ADD_DEVICE) {
      // let USB LED flash
      File = fopen("/sys/class/leds/USB/brightness", "w");
      if(File) {
        fwrite("2", 1, sizeof(char), File);
        fclose(File);
      }

      coldpluggedDevice(' ', ADD_DEVICE);
      sleep(1);

      // set off USB LED
      File = fopen("/sys/class/leds/USB/brightness", "w");
      if(File) {
        fwrite("0", 1, sizeof(char), File);
        fclose(File);
      }
    }
    return;
  }

  File = fopen("/sys/class/leds/USB/brightness", "w");
  if(File) {
    fwrite("2", 1, sizeof(char), File);  // LED blinking
    fclose(File);
  }
  #endif

  // ..................................................................
  // The file "/proc/partitions" is not valid after device removal.
  if(Action == REMOVE_DEVICE) {
    removeMassStorageDevice(DevPath, PortNo);
    reconfigSamba();
  }
  else {
    addMassStorageDevice(MountPath, DevPath, PortNo);
    reconfigSamba();
  }

  #ifdef DEVICE_IS_FSG
  File = fopen("/sys/class/leds/USB/brightness", "w");
  if(File) {
    fwrite("0", 1, sizeof(char), File);   // LED off
    fclose(File);
  }
  #endif
}


// ***********************************************************************
// Handle key events.
// ***********************************************************************
#ifdef KEY_ACTIONS
void pressedKey()
{
  #ifndef DEVICE_IS_VT4
  FILE *File;
  #endif
  char *Command, *p;
  int i;
  char *newmaster,*oldoldmaster,*oldmaster,*combmaster,*shadow,*adminpass;
  char badminpass[256];
  if((Command=getenv("ACTION")) == NULL) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Missing action statement.");
    #endif
    return;
  }

  int Action = REMOVE_DEVICE;
  if(strcmp(Command, "REBOOT") == 0) {
    system("reboot");
    return;
  }

  if(strcmp(Command, "SHUTDOWN") == 0) {
    system("halt");
    return;
  }

  if(strcmp(Command,"USEREVENT") == 0) {
    #ifndef DEVICE_IS_VT4
    File = fopen("/sys/class/leds/SYNC/brightness", "w");
    if (File) {
      fwrite ("2", 1, sizeof(char), File);
      fclose (File);
    }
    #endif
    #ifdef DEVICE_IS_VT4
    system("/sbin/led SYNC 1");
    #endif

    newmaster = readfile(MasterConfName, 0);
    if(newmaster == 0)  return;

    p = newmaster;
    // \nbackup=name#0/1#src#dst#username#pwd#h/d/w/m#day#time#cron
    while((p=strstr(p,"\nbackup=")) != 0) {
	Command = strchr(p+8, '\n');
	if(Command == 0)  Command += strlen(p);
	strtok(p,"#");		// skip name
	strtok(NULL,"#");	// skip enabled
	strtok(NULL,"#");	// skip source
	strtok(NULL,"#");	// skip destination
	strtok(NULL,"#");	// skip user
	strtok(NULL,"#");	// skip pwd
	strtok(NULL,"#");	// skip schedule
	strtok(NULL,"#");	// skip day
	strtok(NULL,"#");	// skip time
	p = Command;
	if ((Command=strtok(NULL,"#"))==0)
		continue;

	if ((Command=strchr(Command,'/'))==0)
		continue;

	// The backup program makes sure that it is not called twice.
	system(Command);
    }

    free(newmaster);
    #ifndef DEVICE_IS_VT4
    File = fopen("/sys/class/leds/SYNC/brightness", "w");
    if (File) {
      fwrite ("0", 1, sizeof(char), File);
      fclose (File);
    }
    #endif
    #ifdef DEVICE_IS_VT4
    system("/sbin/led SYNC 0");
    #endif
    return;
  }

  // ACTION: revert to factory default
  if(strcmp(Command, "DEFAULTS") == 0) {
	for(;;){
		newmaster=readfile("/etc/templates/master.conf", 0);
		if(!newmaster)break;
		oldmaster=readfile(MasterConfName, 0);
		oldoldmaster=oldmaster;
		oldmaster=strstr(oldmaster,"\nfirmware_version=");
		if(oldmaster){
			oldmaster+=strlen("\nfirmware_version=");
			while((*oldmaster!='\n')&&(*oldmaster!=0))oldmaster++;
			if(*oldmaster=='\n')oldmaster++;
		} else break;
		combmaster=(char *)malloc(strlen(newmaster)+strlen(oldmaster)+2);
		strcpy(combmaster,newmaster);
		strcat(combmaster,oldmaster);

		 #ifdef DEVICE_IS_DT2
			if(get_value(combmaster, "\nconfig_setup", badminpass))
				change_cmdline(badminpass);
		 #endif

		writefile(MasterConfName,combmaster);
		free((char *)newmaster);
		free((char *)oldoldmaster);
		// reset admin password
		shadow=readfile("/etc/shadow", 0);
		if(!shadow) break;
		adminpass=strstr(shadow,"\nadmin:");
		if(!adminpass)break;
		adminpass++;
		i=0;
		while((*adminpass!=0)&&(*adminpass!='\n')){
			badminpass[i]=*adminpass;
			i++; adminpass++;
		}
		badminpass[i]=0;
		shadow=replace_string_once(shadow,shadow,badminpass,"admin:$1$$CoERg7ynjYLsj2j4glJ34.:12688:0:99999:7:::");
		writefile("/etc/shadow",shadow);
		sprintf(badminpass, "/usr/local/samba/bin/smbpasswd -s -a %s >/dev/null <<!LIM!\n%s\n%s\n!LIM!",
	  		"admin", "admin", "admin");
  		system(badminpass);

		break;
	}
	system("rm /home/aquota.user");
	system("reboot");
	return;
  }

  if(strncmp(Command, "COLDPLUG", 8) == 0)
    Action = COLDPLUG_DEVICE;
  else if(strncmp(Command, "UNPLUG", 6) != 0) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "Unsupported key action: %s.", Command);
    #endif
    return;
  }


  #ifdef DEVICE_IS_FSG
  // let USB LED flash
  File = fopen("/sys/class/leds/USB/brightness", "w");
  if (File) {
    fwrite ("2", 1, sizeof(char), File);
    fclose (File);
  }

  if(Action == COLDPLUG_DEVICE) {

    coldpluggedDevice(' ', ADD_DEVICE);
    sleep(1);

    // set off USB LED
    File = fopen("/sys/class/leds/USB/brightness", "w");
    if (File) {
      fwrite ("0", 1, sizeof(char), File);
      fclose (File);
    }
  }

  else if(Action == REMOVE_DEVICE) {

    removeMassStorage(0, ' '); // removes all SCSI
    removePrinter(0, ' ');     // remove all printer drivers

    reconfigSamba();
    sleep(1);

    // set off SATA LED
    File = fopen("/sys/class/leds/SATA/brightness", "w");
    if (File) {
      fwrite ("0", 1, sizeof(char), File);
      fclose (File);
    }

    // set on USB LED
    File = fopen("/sys/class/leds/USB/brightness", "w");
    if (File) {
      fwrite ("1", 1, sizeof(char), File);
      fclose (File);
    }
  }
  #endif


  #ifdef DEVICE_IS_DT2
  if(Action != REMOVE_DEVICE)
    return;

  // let USB LED flash
  char PortNo = *(Command+strlen(Command)-1);
  if(PortNo != '2')
    Command = "/sys/class/leds/USB1/brightness";
  else
    Command = "/sys/class/leds/USB2/brightness";

  File = fopen(Command, "w");
  if (File) {
    fwrite ("2", 1, sizeof(char), File);
    fclose (File);
  }

  // Terminating and restarting samba solves problems with unmounting,
  // but of course it would raise new problems.
//  system("/etc/init.d/samba stop");
//  sync();
//  usleep(20000);
  if(coldpluggedDevice(PortNo, REMOVE_DEVICE) == 0) {

    File = fopen(Command, "w");
    if (File) {
      fwrite ("1", 1, sizeof(char), File);
      fclose (File);
    }
  }
//  system("/etc/init.d/samba start");
  #endif


  #ifdef DEVICE_IS_VT4
  if(Action != REMOVE_DEVICE)
    return;

  coldpluggedDevice(*(Command+strlen(Command)-1), REMOVE_DEVICE);
  #endif

}
#endif


// *************************************************************************
// **********                                                     **********
// **********                   MAIN FUNCTION                     **********
// **********                                                     **********
// *************************************************************************

int main(int argc, char** argv)
{
  if(argc != 2) {
    #ifdef ERROR_LOG
      syslog(LOG_ERR, "%s.", "Missing device class");
    #endif
    return 1;          // no action reported ?
  }
  #ifdef INFO_LOG
    syslog(LOG_NOTICE, "********* Plug event \"%s\". *********", argv[1]);
  #endif

  if(strcmp(argv[1], "block") == 0)
    hotpluggedBlockDevice();
  else if(strcmp(argv[1], "usb") == 0)
    hotpluggedPrinter();
  else if(strcmp(argv[1], "cleanup") == 0) // clean up all USB things
    cleanupFolders();
  else if(strcmp(argv[1], "coldusb") == 0) // coldplugged USB device
    coldpluggedDevice(' ', ADD_DEVICE);
  #ifdef KEY_ACTIONS
  else if(strcmp(argv[1], "key") == 0) // pressed button
    pressedKey();
  #endif
  else {
  #ifdef INFO_LOG
    syslog(LOG_ERR, "Unsupported device class \"%s\".", argv[1]);
  #endif
    return 1;
  }

  return 0;
}
